import random
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
from collections import deque
import sys
import pdb
import pylab

#When a creature makes a decision, it will return an object of type Action
#that specifies the desired behavior. Don't worry about the details. If you're
#just opening this code for the first time, follow the directions in the assignment.
#
#You are not expected to read all of this code and understand it (though you're welcome
#to if you'd like).

#You should not modify the Action class until you have finished all but the
#optional part of this assignment


class Action(object):
    valid_directional_action_types = ["attack", "replicate", "move"]
    valid_stationary_action_types = ["pass", "forage", "die"]
    valid_action_types = valid_directional_action_types + valid_stationary_action_types
    
    valid_directions = set(["top", "bottom", "left", "right"])
    
    #action_type and direction are strings, and must obey the rules given above
    def __init__(self, action_type, direction):
        if action_type not in self.valid_action_types:
            raise Exception("Invalid action_type: %s", action_type)
        
        if (action_type in self.valid_stationary_action_types
            and direction != None):
                raise Exception("Direction %s cannot be specified for action_type %s" % (
                                   direction, action_type))

        if (action_type in self.valid_directional_action_types
            and direction not in self.valid_directions):
                raise Exception("Invalid direction: %s (For action_type %s)" % (
                                   direction, action_type))
                        
                                                                        
        self.action_type = action_type
        self.direction = direction

##There are four 'creature' types in huglife:
#1. Plip
#2. Clorus
#3. EmptySquare
#4. ImpassibleSquare

#Only Plips and Cloruses are actually creatures.
#
#Empty squares are used to represent empty spaces in the grid of the world, since
#the design of the world simulator assumes there is something in every square (i.e.
#you're not allowed to have objects of type None in the world grid). 
#
#Impassible squares are used to represent spaces outside the grid universe. You
#won't have to use it, but there's a function called get_creature(x, y). If x and
#y are out of bounds, an ImpassibleSquare is returned.
#
#You could also in principle use this class to create impassible walls in the 
#middle of the universe.

#this class should not be modified
class ImpassibleSquare(object):
    def __init__(self):
        self.name = "blocked"
        #R=0, G=0, B=0, this is black.
        self.basecolor = [0.0, 0.0, 0.0] 

    def color(self):
        return self.basecolor        
                        
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return True
        return False        

#this class should not be modified
class EmptySquare(object):
    def __init__(self):
        self.name = "empty"
        #R=1, G=1, B=1, this is white.
        self.basecolor = [1.0, 1.0, 1.0] 

    def color(self):
        return self.basecolor

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return True
        return False       


#When you've completed the Plip class, they will gain food through foraging,
#standing still unless there is a threat.

#See the final required part of the assignment for a more detailed specification
#of desired Plip behavior.

class Plip(object):
    #costs for actions
    move_cost = 0.1
    pass_cost = 0.01
    replication_factor = 0.5

    #gains for food acquisition    
    forage_gain = 0.1    
    max_energy = 2
    
    #parameters that affect decision making
    move_chance = 0.25
    replication_threshold = 1
    
    #minimum energy before death    
    death_threshold = 0
    
    def __init__(self, energy):
        self.name = "plip"
        self.energy = energy
        
        #R=0, G=1.0, B=0, this is green.
        self.basecolor = np.array([0.0, 1.0, 0.0]) 

    #returns a color for this plip, with brightness proportional to its energy
    def color(self):
        #brightness_level is a number between 0.3333 and 1. There is a linear
        #mapping from energies between 0 and 2 to the range 0.333 and 1.
        brightness_level = (self.energy + 1) / (1 + self.max_energy)
        return self.basecolor * brightness_level                    
        
    ####YOU DO NOT NEED TO MODIFY ANYTHING IN THE PLIP CLASS ABOVE THIS LINE
        
    #Energy is reduced by an amount equal to self.move_cost
    def spend_move_energy(self):
        self.energy = self.energy - 0

    #Energy is increased by self.forage_gain
    #If energy exceeds self.max_energy, it should be set equal to self.max_energy
    def gain_energy_from_foraging(self):
        self.energy = self.energy + 0
        
    #returns a new plip with a energy equal to parent's energy times
    #replication factor.
    
    #side effect: parent also ends up with energy times replication factor    
    def replicate(self):   
        #the code that is here right now is bad! the baby plip has just as much
        #energy as the parent plip, and the parent plip doesn't lose any energy
        baby_plip = Plip(self.energy)
        return baby_plip 
              
                              
    #uses neighbors_by_occupant to decide what to do
    #neighbors_by_occupant is a dictionary that maps from each of the
    #four creature types to a list of directions in which creatures of that type
    #can be found. 
    
    #For example, if there is a clorus above, an empty square below, and a
    #pair of plips to the left and right, then the dictionary would be:
    
    #neighbors_by_occupant["clorus"] == ["top"]
    #neighbors_by_occupant["plip"] == ["left", "right"]
    #neighbors_by_occupant["empty"] == ["bottom"]
    #neighbors_by_occupant["blocked"] == []
    
    #In this case, since none of the four squares are blocked, the list corresponding
    #to the "blocked" key is empty.   
    
    #choose_action should return an Action with action_type equal to one of
    #the 6 valid actions described in the assignment spec and in the action class
    #above, and with direction one of the four valid directions if the action_type
    #involves movement (attack, move, replicate), and should be None if the
    #action_type does not involve movement (forage, die, pass)
    def choose_action(self, neighbors_by_occupant):
        possible_directions = ["top", "bottom", "left", "right"]
        
        #random.choice is a neat function that returns one random item from a list
        move_direction = random.choice(possible_directions)
        
        return Action("move", move_direction)

#When you've completed the Clorus class, they will roam the world looking for
#plips to eat. If they find a plip and they are not surrounded, they will eat that
#plip. If they are ever surrounded by 4 plips, they will pass.
            
#See the final required part of the assignment for a more detailed specification
#of desired Clorus behavior.            
            
class Clorus(object):
    #costs for actions
    move_cost = 0.025
    replication_factor = 0.5
    pass_cost = 0.125
        
    #food acquisition parameters
    eat_efficiency = 0.9    
    max_energy = 2
    
    #parameters that affect decisions
    replication_threshold = 1

    #minimum energy before death                
    death_threshold = 0
                                
    def __init__(self, energy):
        self.energy = energy
        self.name = "clorus"
        self.basecolor = np.array([1.0, 0.0, 0.0]) #R=1.0, G=0, B=0, this is red.

    def color(self):
        return self.basecolor * (self.energy + 1) / (1 + self.max_energy)         
 
     ####YOU DO NOT NEED TO MODIFY ANYTHING IN THE CLORUS CLASS ABOVE THIS LINE

 
    #decreases energy by an amount equal to self.move_cost
    def spend_move_energy(self):
        self.energy = self.energy - 0

    #decreases energy by an amount equal to self.pass_cost        
    def spend_pass_energy(self):
        self.energy = self.energy - 0
        
    #increases energy by an amount equal to the victim's energy times
    #self.eat_efficiency. if energy is about max_energy, then set energy
    #to max_energy
    def gain_energy_from_eating(self, eaten_creature):
        self.energy = self.energy + 0
     
    #returns a new clorus with a energy equal to parent's energy times
    #replication factor.
    
    #side effect: parent also ends up with energy times replication factor
    def replicate(self):
        baby_clorus = Clorus(self.energy)
        return baby_clorus
        
    def choose_action(self, neighbors_by_occupant):                
        possible_directions = ["top", "bottom", "left", "right"]
        
        #random.choice is a neat function that returns one random item from a list
        move_direction = random.choice(possible_directions)
        
        return Action("move", move_direction)
        
        
        
        
        
        
        
        
        
        
        
        
        
        
            
#####################################################################
#
#
#  To complete this assignment, you do not need to modify anything
#  below this line!
#
# However, you are welcome to try experimenting with the code below.
#####################################################################


# The basic flow of the simulator (given in the Grid class)
# is as follows:

# The universe proceeds one cycle at a time. A cycle consists of every
# creature in the universe getting a chance to move exactly once. Each move
# is also called a tic.
#
# When a cycle begins, a creature_queue is created, and each creature is put
# into the queue in some arbitrary order. When a creature reaches the front of
# the queue, it is given information about its four neighbors, and must return
# an object of type Action -- unless it has less than death_threshold energy,
# in which case the simulator picks the Action "die" without letting the creature
# make a decision.

# If the action is "pass", "move", "attack", "forage", or "replicate", the world
# simulator calls the spend_pass_energy, spend_move_energy, gain_energy_from_eating,
# gain_energy_from_foraging, or replicate methods respectively. These methods
# are responsible for adjusting the internal state of the creature in response
# to the gains or costs -- except replicate which should also return a new creature of
# the replicating type.

# The simulator also adjusts the state of the world as follows: 
# If the action was "move", the deciding creature is placed into the appropriate square.
# If the action was "attack", the attacked creature is removed from the world, 
#    and the attacker takes its place on the grid.
# If the action was "replicate", the baby creature is placed into the empty square
#    specified.
# The simulator does not do anything to the state of the world when "move" or
#    "pass" are chosen (though the creatures themselves may incur changes
#    from their gain_energy_from_foraging or spend_pass_energy methods).

# After a creature completes its action, it is placed at the back of the queue, and
# will take a move during the next cycle. If an action destroys a creature, it
# is removed from the queue entirely. New creatures are also placed at the back
# of the queue, ready to take their first action next turn (as long as they aren't
# eaten before they get a chance).



#create an empty grid unpythonically
#a better approach is to use a list comprehension
#but we have not learned this yet
def create_nxn_empty_grid(N):
    grid = []
    for i in range(N):
        row = []
        grid.append(row)
        for j in range(N):
            row.append(EmptySquare())
    return grid

#Returns True if the grid point supplied is empty
#Returns False if the grid point supplied is not empty
#Happens to return true if supplied argument is not actually a grid point
#but this is just out of laziness
def is_empty(grid_point):
    return isinstance(grid_point, type(EmptySquare()))

def is_clorus(grid_point):
    return isinstance(grid_point, type(Clorus(0)))

def is_plip(grid_point):
    return isinstance(grid_point, type(Plip(0)))

def is_impassible(grid_point):
    return isinstance(grid_point, type(ImpassibleSquare()))



class Grid:
    
    #dictionary of dictionaries
    def __init__(self, N):
        self.grid_points = create_nxn_empty_grid(N)
        self.N = N
        self.clorus_pop = 0
        self.plip_pop = 0
        self.creature_queue = deque()
        self.actions_this_cycle = 0
        self.population_history = []

                        
    def in_bounds(self, x, y):
        if x < 0 or y < 0:
            return False
        
        if x >= self.N or y >= self.N:
            return False
            
        return True    
    
    def get_creature(self, x, y):
        #things are stored in memory by row and column instead of x and y
        #annoying, but true!
        row = y
        column = x
        
        if self.in_bounds(x, y):
            return self.grid_points[row][column]
        else:
            return ImpassibleSquare()

    def place_creature(self, x, y, creature):
        if self.in_bounds(x, y):
            old_creature = self.get_creature(x, y)
            if (x, y) in self.creature_queue:
                self.creature_queue.remove((x, y))                
                self.actions_this_cycle += 1
            
            if old_creature.name == "plip":
                self.plip_pop -= 1
                
            if old_creature.name == "clorus":
                self.clorus_pop -= 1
            
            if creature.name == "plip":
                self.plip_pop += 1
                self.creature_queue.append((x, y))

                                
            if creature.name == "clorus":
                self.clorus_pop +=1 
                self.creature_queue.append((x, y))                

            #things are stored in memory by row and column instead of x and y
            #annoying, but true!                                
            row = y
            column = x
            self.grid_points[row][column] = creature
            


    #returns true if there is at least one creature left in the universe
    def life_exists(self):
        if (self.clorus_pop == 0) and (self.plip_pop == 0):
            return False
        return True


    #the second maps creature type to a list of directions
    def neighbors_by_occupant_type(self, x, y):
        neighbors = {}
        neighbors["empty"] = []
        neighbors["blocked"] = []
        neighbors["clorus"] = []
        neighbors["plip"] = []
        top_name = self.get_creature(x, y+1).name
        bottom_name = self.get_creature(x, y-1).name
        left_name = self.get_creature(x-1, y).name
        right_name = self.get_creature(x+1, y).name
        neighbors[top_name].append("top")
        neighbors[bottom_name].append("bottom")
        neighbors[left_name].append("left")
        neighbors[right_name].append("right")
        return neighbors        

    @staticmethod
    def direction_name_to_xy(x, y, direction):
        if direction == "top":
            return x, y+1
        if direction == "bottom":
            return x, y-1
        if direction == "left":
            return x-1, y
        if direction == "right":
            return x+1, y
        
        raise Exception("Invalid direction name: %s, must be 'top', 'bottom', 'left', or 'right'" % direction)        
        return None

    def handle_action(self, x, y, creature, action):        
        if action.action_type == "move":
            #print("moving from %d, %d to %d, %d"% (x, y, newX, newY))
            newX, newY = Grid.direction_name_to_xy(x, y, action.direction)
            existing_creature = self.get_creature(newX, newY)
            if not is_empty(existing_creature):
                exception_message = ("Invalid action by %s in position (%d, %d). Tried to move " +
                                    "into position (%d, %d), already occupied by %s") % (
                                    creature.name, x, y, newX, newY, existing_creature.name)
                                    
                raise Exception(exception_message)                

            self.place_creature(x, y, EmptySquare())
            self.place_creature(newX, newY, creature)
            creature.spend_move_energy()
                        
        if action.action_type == "replicate":
            newX, newY = Grid.direction_name_to_xy(x, y, action.direction)
            new_creature = creature.replicate()
            existing_creature = self.get_creature(newX, newY)
            if not is_empty(existing_creature):
                exception_message = ("Invalid action by %s in position (%d, %d). Tried to replicate " +
                                    "into position (%d, %d), already occupied by %s") % (
                                    creature.name, x, y, newX, newY, existing_creature.name)
                                    
                raise Exception(exception_message)               
                
            self.place_creature(x, y, creature)
            self.place_creature(newX, newY, new_creature)
                
        if action.action_type == "forage":
            creature.gain_energy_from_foraging()
            if not is_plip(creature):
                exception_message = ("Invalid action by %s in position (%d, %d). Tried to forage, but " +
                                    "only plips may forage") % (
                                    creature.name, x, y)
                                    
                raise Exception(exception_message)  
                                       
            if creature.energy > creature.max_energy:
                exception_message = ("After foraging, Plip in position (%d, %d) has energy %.2f, greater than" +
                                    " max allowed energy: %.2f") % (x, y, self.energy, self.max_energy)
                                    
                raise Exception(exception_message)                  

            self.place_creature(x, y, creature)


        if action.action_type == "pass":
            creature.spend_pass_energy()      
            self.place_creature(x, y, creature)
               
        if action.action_type == "die":
            self.place_creature(x, y, EmptySquare())            
                                    
        if action.action_type == "attack":
            newX, newY = Grid.direction_name_to_xy(x, y, action.direction)
            eaten_creature = self.get_creature(newX, newY)
            creature.gain_energy_from_eating(eaten_creature)

            if not is_clorus(creature) or not is_plip(eaten_creature):
                exception_message = ("Invalid action by %s in position (%d, %d). Tried to eat " +
                                    "%s in position (%d, %d).\nCloruses may eat plips." +
                                    "No other eating actions are allowed") % (
                                    creature.name, x, y, newX, newY, existing_creature.name)
                                    
                raise Exception(exception_message)                 
                                                     
            if creature.energy > creature.max_energy:
                exception_message = ("After eating Plip in position (%d, %d), Clorus that was in position " +
                                     "(%d, %d) has energy %.2f, greater than" +
                                     " max allowed energy: %.2f") % (newX, newY, x, y, self.energy, self.max_energy)
                                    
                raise Exception(exception_message)                   
                
            self.place_creature(newX, newY, creature)
            self.place_creature(x, y, EmptySquare())
                                                                                        
    def single_cycle(self):
        self.population_history.append(self.clorus_pop + self.plip_pop)

        if not self.life_exists():
            if random.random() > 0.01:
                #x = random.randint(0, self.N)
                #y = 0
                #self.place_creature(x, y, Plip(0.5))
                #print("spontaneous generation!")
                return
            
        self.actions_this_cycle = 0
        actions_to_take = self.clorus_pop + self.plip_pop
            
        while self.actions_this_cycle < actions_to_take:
            self.single_tic()        
    
    def single_tic(self):
        if self.life_exists():
            x, y = self.creature_queue.popleft()
            self.actions_this_cycle += 1
            
            creature = self.get_creature(x, y)
            neighbors = self.neighbors_by_occupant_type(x, y)
            
            #if a creature has less energy than its given death threshold
            #it is forced to choose the action 'die'.
            if creature.energy < creature.death_threshold:
                action = Action("die", None)
            else:
                action = creature.choose_action(neighbors)
            self.handle_action(x, y, creature, action)            
            

    #draws the state of the world as a single static frame
    #not suitable for real time animation!
    def draw_world(self):
        plt.clf()
        fig = plt.figure(1)
        
        colors = np.zeros(shape=(self.N, self.N, 3))
        
        for x in range(self.N):
            for y in range(self.N):
                xy_creature = self.get_creature(x, y)
                #set red, green, and blue
                xy_creature_color = xy_creature.color()                
                colors[x][y][0] = xy_creature_color[0]
                colors[x][y][1] = xy_creature_color[1]
                colors[x][y][2] = xy_creature_color[2]



        plt.imshow(np.transpose(colors, (1, 0, 2)), interpolation="nearest", origin='lower')
        fig.canvas.draw()
        plt.draw()
        plt.show()

    def step_by_step_animation(self, frames):
        import pylab
        for i in range(frames):
            self.single_tic()
            self.draw_world()                             
            pylab.pause(0.1)            
            raw_input("press enter to continue")

    def animate(self, i):            
    
        self.single_cycle()
        
        #if the world is dead and it was last turn as well, then abort drawing
        if i > 0 and self.population_history[i-1] == 0 and not self.life_exists():
            return [self.animation_frame]   
                 

        
        colors = self.animation_frame.get_array()
        for x in range(self.N):
            for y in range(self.N):
                xy_creature = self.get_creature(x, y)
                #set red, green, and blue
                xy_creature_color = xy_creature.color()
                colors[x][y][0] = xy_creature_color[0]
                colors[x][y][1] = xy_creature_color[1]
                colors[x][y][2] = xy_creature_color[2]

        self.animation_frame.set_array(np.transpose(colors, (1, 0, 2)))        

                
        return [self.animation_frame]        

    #simulates a number of timesteps given by frames
    def begin_animated_simulation(self, time_between_frames, frame_count):
        all_white_matrix = np.ones(shape=(self.N, self.N, 3))
        self.animation_frame = plt.imshow(all_white_matrix, interpolation="nearest", origin='lower')        
        self.fig = plt.figure(1)
        ax = self.fig.add_axes([0, 0, self.N, self.N], xticks=[], yticks=[], frameon=False)        

        #animation starts. matplotlib doesn't seem to behave very well when you tell it not
        #to loop forever, so I've disabled the repeat=False parameter
        anim = animation.FuncAnimation(self.fig, self.animate, frames=frame_count,
                               interval=time_between_frames, blit=True)#, repeat=False)         
             
        #if you have the appropriate programs installed on your computer, you can
        #generate mp4 files by uncommenting this line.                  
        #anim.save('some_file.mp4', fps=30, extra_args=['-vcodec', 'libx264'])
        plt.show()        